/*
contiene tutte le routine per la gestione a basso ed alto livello dell'I2C
*/

/*I2cLowService ***********************************************************************
Per la gestione a basso livello della comunicazione I2c
*/

void I2cLowService (void)
{
	if (I2cBusCollFlag || SSPCON2bits.ACKSTAT)
	{	// c'e' stata una collisione sul bus o un NACK -> annnulla sequenza
		I2cBusCollFlag = 0;			 // reset stato errore
		I2cStat = FINE;				 // stato attuale = FINE sequenza
	   StopI2C();						 // invia stop
	   // non azzera il contatore byte Tx e Rx per ritentare scambio al prossimo giro	  
	}
	
	else
	{	
	// esegue la sequenza I2c stabilita per ogni byte nel record puntato
	
		switch (I2cStat)
		{
			case (START):	// e' stata eseguita una sequenza di start		
			   if (I2c[I2cDevPtr].Flag.Tx > 0)// c'e' qualcosa da trasmettere ?
			   {
 	           I2cStat = WRITE;             // aggiorna stato I2c attuale
 	           SSPBUF = I2cAdr[I2cDevPtr];  // scrive indirizzo con flag R/W = write
				}
				else     // se non ha dati da trasmettere ne ha sicuramente da ricevere
				{
   	        I2cStat = READ;              // aggiorna stato I2c attuale	
   	        SSPBUF = I2cAdr[I2cDevPtr]+1;// scrive indirizzo con flag R/W = read
   			} 
				break;
			
			
			case (WRITE):
				/* invia il byte Nesimo
					controlla se sono stati inviati tutti i byte
					controlla se ci sono byte da ricevere
					altrimenti chiude la sequenza
				*/
				SSPBUF = I2c[I2cDevPtr].TxBuff[Ptr.Tx];    // invio primo byte 
				Ptr.Tx ++;                                 // byte Tx successivo 
				if (Ptr.Tx >= I2c[I2cDevPtr].Flag.Tx)      // ha finito Tx
				{
   				if (I2c[I2cDevPtr].Flag.Rx > 0)  // ha qualcosa da ricevere ?
   				{
   					I2cStat = RSTART;             // inizia sequenza di ricezione	
   				}
   				else										// non ha niente da ricevere
   				{
      				I2cStat = STOP;               // invia stop regolare 
     				}
    			}
    			else											// ha ancora byte da trasmettere
    			{
    				I2cStat = WRITE;          			// inizia sequenza invio byte(s) successivo(i)	
    			}
				break;
				
				
			case (READ):
				SSPCON2bits.RCEN = 1;           				// avvia ricezione di un byte
				Ptr.Rx ++;                        			// byte Rx successivo 
				if (Ptr.Rx >= I2c[I2cDevPtr].Flag.Rx)     // ha finito Rx ?
				{
   				I2cStat = NACK;				   			// SI, invia NACK (fine Rx)
 				}
				else
				{
					I2cStat = ACK;									// NO. invia ACK (prosegue Rx)
				}			
				break;
				
				
			case (ACK):
				I2c[I2cDevPtr].RxBuff[Ptr.Rx-1] = SSPBUF;	// memorizza byte Nesimo ricevuto 
																		// il Ptr era stato gia' incrementato 
				I2cStat = READ;									// altri byte da ricevere
				AckI2C();
				break;
			
				
			case (NACK):
				I2c[I2cDevPtr].RxBuff[Ptr.Rx-1] = SSPBUF;	// memorizza byte Nesimo ricevuto 
																		// il Ptr era stato gia' incrementato 
				I2cStat = STOP;									// ricevuto ultimo byte
			  	NotAckI2C();			  			 	
				break;
				
				
			case (RSTART):
				// reinizializza bus senza rilasciarlo
				I2cStat = ADRR;              // al prossimo giro invia inizio ricezione
				RestartI2C();
				break;
				
				
			case (ADRR):
				I2cStat = READ;              // al prossimo giro inizia la ricezione
				SSPBUF = I2cAdr[I2cDevPtr]+1;// scrive indirizzo con flag R/W = read
  				break;
		
		
			case (STOP):			
    			I2cStat = FINE;             // aggiorna stato I2c attuale
      		I2c[I2cDevPtr].Flag.Rx = 0; // non ci sono altri byte da Tx o Rx le routine a
	   		I2c[I2cDevPtr].Flag.Tx = 0; // livello piu' alto possono scambiare altri dati
	   		StopI2C();					    // invia stop
				break;
				
				
			case (FINE):
				I2cBusyFlag = 0;	  	       // La comunicazione I2c e' finita
				break;
			
				
			default:
				break;
				
		}	// end switch
	}	// end else
					
} // I2cLowService
/**************************************************************************************/


/*I2cHighService ***********************************************************************
Per la gestione ad alto livello della comunicazione I2c
*/

void I2cHighService (void)
{
	/*
	 inizializza sequenza I2c azzerando flag, contatori e buffer
	 lancia il primo start del protocollo
	 il resto dello scambio del pacchetto sara' eseguito tramite interrupt
	*/
	Ptr.Tx = 0;				// Azzera puntatore buffer dati da trasmettere 
	Ptr.Rx = 0;				// Azzera puntatore buffer dati da ricevere
	I2c[I2cDevPtr].RxBuff[0] = 0; // Azzera buffer ricezione
	I2c[I2cDevPtr].RxBuff[1] = 0; // Azzera buffer ricezione
	I2cEventFlag = 0;		// Evento I2c sulla porta SSP
	I2cBusCollFlag = 0;	// Collisione su bus I2c
	I2cBusyFlag = 1;	  	// E' in corso una comunicazione I2c
	I2cStat = START;		// Contatore stati I2c
	
	StartI2C();				// inizia scambio I2c con il primo stato
	
} // I2cHighService
/**************************************************************************************/
