/*	
viene incluso in "dino.c", contiene:
	definizioni
	prototipi funzioni
	settings
	descrizioni
*/


/* Definizioni globali++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#define PLL 1

//----- Status bits
#ifdef PLL  // 40MHz
	#pragma config OSC = HSPLL			//  10 x 4 = 40MHZ
#else
	#pragma config OSC = HS				//	 20Mhz
#endif

#pragma config WDT = OFF
#pragma config LVP = OFF
// #pragma config PWRT = ON   non compatibile con ICD2
#pragma config XINST = ON

//----- Definizione porte
#define EncoderRint INTCON3bits.INT1IF		// interrupt flag encoder SX
#define EncoderLint INTCON3bits.INT2IF		// interrupt flag encoder DX
#define KeypadInt INTCONbits.INT0IF			// interrupt flag tasto premuto

												// B0 Interrupt I/O exp		 I INT0
#define EncoderRpulse PORTBbits.RB1	// B1 Encoder SX,ticks      I INT1
#define EncoderLpulse PORTBbits.RB2	// B2 Encoder DX,ticks      I INT2
											   // B3
#define EncoderRdir PORTBbits.RB4  	// B4 Encoder SX,direzione  I
#define EncoderLdir PORTBbits.RB5  	// B5 Encoder DX,direzione  I	
											  	// B6 PGC (ICD2)				 
												// B7 PGD (ICD2)			  	  
												// C0
												// C1 PWMmotR         	    O PWM   
												// C2 PWMmotL               O PWM
												// C3 SCL
												// C4 SDA
#define Buzzer PORTCbits.RC5     	// C5 Buzzer                O
#define MotEnable PORTDbits.RD0  	// D0 Enable H-bridge       O
#define SoundIn PORTDbits.RD1			// D1 Tone  decoder 4KHz	 I
#define BumperDx PORTDbits.RD2   	// D2 Bumper destro         I
#define BumperSx PORTDbits.RD3   	// D3 Bumper sinistro       I 
#define LedRosso PORTDbits.RD4   	// D4 LedRosso (temp)       O
#define LedVerde PORTDbits.RD5   	// D5 LedVerde (temp)       O
#define LedGiallo PORTDbits.RD6  	// D6 LedGiallo (temp)      O
#define LcdEN PORTDbits.RD7			// D7 Lcd Enable      		 O

struct Bits{
	unsigned bit0:1;
	unsigned bit1:1;
	unsigned bit2:1;
	unsigned bit3:1;
	unsigned bit4:1;
	unsigned bit5:1;
	unsigned bit6:1;
	unsigned bit7:1;
};

struct Bits VARbits1;

struct Bits VARbits2;

struct Bits VARbits3;
	
struct Bits VARbits4;

//----- Flags
#define  PidTick VARbits1.bit0         // timer routine PID, impostato dalla ISR 
#define  Space2RunFlag VARbits1.bit1   // Abilita la routine di controllo dello spazio da percorrere
#define  FlagBumpers VARbits1.bit2     // Abilita la routine controllo collisioni
#define  FlagStop VARbits1.bit3        // Flag per stop motori
#define  FlagAD VARbits1.bit4    	   // Flag per controllo stato routine conversione AD
#define  FlagEye VARbits1.bit5		   // Flag abilitazione routine Eye
#define  FlagLight VARbits1.bit6       // Flag abilitazione routine Light
#define  FlagSound VARbits1.bit7       // Flag abilitazione routine Sound
#define  FlagTargetLight VARbits2.bit0 // Flag abilitazione sensori luce dopo "obiettivo raggiunto"
#define  FlagTargetSound VARbits2.bit1 // Flag abilitazione sensori suono dopo "obiettivo raggiunto"
#define  FlagTargetGas VARbits2.bit2   // Flag abilitazione sensore gas dopo "obiettivo raggiunto"
#define  RandomBit VARbits2.bit3 	   // Cambia da 1 o 0 ogni millisecondo
#define  FlagGas VARbits3.bit2			// Flag abilitazione routine gas
#define  FlagTime VARbits3.bit4			// Flag abilitazione routine Temporizzazioni


//----- Variabili
int i = 0; // contatore generico
int j = 0; // contatore generico

// le variabili long seguenti contengono impulsi dell'encoder
long EncoderRcount = 0;      // impulsi encoder destro
long EncoderLcount = 0;      // impulsi encoder sinistro
long EncoderRcountPrev =0;   // misura precedente per il computo della velocita'
long EncoderLcountPrev =0;   // spazio percorso = valore attuale - valore precedente

long Space2RunR = 0;         // spazio da percorrere 
long Space2RunL = 0;         // 
long Space2RunRstart = 0;    // valore di partenza per il calcolo della strada da percorrere
long Space2RunLstart = 0;    //

int SpeedR = 0;              // velocita' attuale ruota destra
int SpeedL = 0;              // velocita' attuale ruota sinistra
int DesSpeedR = 0;           // velocita' desiderata ruota destra
int DesSpeedL = 0;           // velocita' desiderata ruota sinistra
int IntR = 0;                // accumulatore errore integrale del PID (motore destro)
int IntL = 0;                // accumulatore errore integrale del PID (motore sinistro)
int ErroreRprev = 0; // errore precedente, serve per il calcolo dell'errore differenziale del PID (motore destro)
int ErroreLprev = 0; // errore precedente, serve per il calcolo dell'errore differenziale del PID (motore sinistro)
int TimerStop;               // timer fino a 30 Sec per routine Stop motori
int TimerBumpers;            // timer per routine Bumpers
int TimerBeep = 0;			  // timer per routine Beep
int TimerSensor = 0;			  // timer per disabilitazione sensori
int TimerDeadCorner = 0;	  // timer per routine angoli morti

unsigned char DeltaT;         // tempo trascorso, per il computo della velocita'
char DeadCorner = 0;          // conta le collisioni sullo stesso lato per uscire da un angolo morto
#define EyeLlVal LcdVar[1][0] // valore attuale del rivelatore IR laterale sinistro
#define EyeLfVal LcdVar[1][1] // valore attuale del rivelatore IR frontale sinistro
#define EyeRfVal LcdVar[1][2] // valore attuale del rivelatore IR frontale destro
#define EyeRlVal LcdVar[1][3] // valore attuale del rivelatore IR laterale destro

#define FrLVal LcdVar[3][0]   // valore attuale della Fotoresistenza Sinistra
#define FrCVal LcdVar[3][1]   // valore attuale della Fotoresistenza Centrale
#define FrRVal LcdVar[3][2]   // valore attuale della Fotoresistenza Destra
#define GasVal LcdVar[3][3] 	// valore attuale del Sensore di gas
char DebounceLight =0 ;		   // contatore per debounce target ok luce
char DebounceSound =0 ;		   // contatore per debounce target ok suono
char BeepCount = 0;			   // indica quanti beep verranno suonati

char TimerAD = 0;             // timer per routine AD
char PathSeq [11]; // contiene la sequenza dei movimenti da fare (n-1 + EOL)
char PathSeqPointer; // puntatore dell'array sequenza


//----- Definizione parametri movimento
/* diametro ruota = 58mm -> circonferenza = 182,212mm
   l'encoder e' su uno degli assi del gruppo riduttore e gira 20 volte piu' della ruota
   questo ha 40 finestrelle che danno un interrupt ognuna
   quindi 40 x 20 = 800 tick per giro -> risoluzione 0,227765 mm
*/
#define stepR  228   //  in millesimi di millimetro
#define stepL  228   //  i valori leggermente diversi correggono le piccole differenze tra le due ruote
#define constSpeed 350 // velocita' normale, da verificare se il PID ha ancora sufficiente controllo
#define lowSpeed   200 // velocita' ridotta per approccio ostacoli
#define manSpeedFwd   200 // velocita' delle ruote durante le manovre
#define manSpeedRew  -200 // tenendola bassa la precisione e' maggiore

#define kp LcdVar[5][0]    // parametro memorizzato in EEPROM interna
#define ki LcdVar[5][1]    // parametro memorizzato in EEPROM interna 
#define kd LcdVar[5][2]    // parametro memorizzato in EEPROM interna 
#define En LcdVar[5][3]    // parametro memorizzato in EEPROM interna enable motori

                                   
/* 
	parametri per il moto in vicinaza degli ostacoli
   sono prese in considerazione due soglie di allarme per la misura dei sensori IR
	sotto LowDist il bot comincia a rallentare
	sotto MinDist gira in senso opposto al lato dell'ostacolo rivelato
*/

// luce
#define Light1 LcdVar[4][0]    // parametro memorizzato in EEPROM interna
#define Light2 LcdVar[4][1]    // parametro memorizzato in EEPROM interna
// queste soglie sono valide per il campo di gara del Von Neumann
#define DebounceLightCount 20 // soglia conteggi target ok luce
#define TimeDisSensLight TimeMotStopTarget + 6000 //tempo sensori off oltre stop motori (in mS)

// gas
#define GasOn LcdVar[4][3]     // parametro memorizzato in EEPROM interna
#define TimeDisSensGas TimeMotStopTarget + 15000 //tempo sensori off oltre stop motori (in mS)

// suono
#define DebounceSoundCount 10 // soglia conteggi target ok suono
#define TimeDisSensSound TimeMotStopTarget + 6000 //tempo sensori off oltre stop motori (in mS)

#define BeepTime 200				// durata/2 in ms del periodo di beep
#define TimeMotStopTarget 1500 // motori fermi per 1,5 sec raggiunto l'obiettivo
// secondo il regolamento dovrebbe fermarsi per 4 sec ma... non lo rispetta nessuno										  						  

#define MinDist 0x60 // 6cm distanza di guardia dei sensori, sotto questa distanza gira 
#define LowDist 0x26 // 15cm distanza di guardia dei sensori, sotto questa distanza rallenta
#define TargetDist 0x14 // 22 cm distanza di "bersaglio individuato"
// il valore esadecimale fa riferimento alla tabella sotto

// ATTENZIONE: a tensione minore corrisponde distanza maggiore

/* sotto i 5cm gira di 10 nella direzione opposta a quella del sensore che lo ha rilevato.
----------------------------------------------------------------------------------------
lettura sensori  IR Sharp GP2D120
distanza dal paraurti

distanza    cm:25 24 23 22 21 20 19 18 17 16 15 14 13 12 11 10 09 08 07 06 05 04 03 02
lettura AD hex:10 11 13 14 16 18 1a 1e 20 24 26 2b 30 33 36 3b 3f 46 4e 5a 66 77 8d a2   

sotto i 2cm la  tensione ricomincia a  scendere.
----------------------------------------------------------------------------------------*/

/*Porte analogiche, ADCON0
bit 5-2 CHS2:CHS0: Analog Channel Select bits
0000 = Channel 0 (AN0)   A0
0001 = Channel 1 (AN1)   A1
0010 = Channel 2 (AN2)   A2
0011 = Channel 3 (AN3)   A3
0100 = Channel 4 (AN4)   A5
0101 = Channel 5 (AN5)   E0
0110 = Channel 6 (AN6)   E1
0111 = Channel 7 (AN7)   E2
*/
#define FrR   	 0b00000000       // AN0 FotoResistenza destra         I analog
#define FrL     0b00000100       // AN1 FotoResistenza sinistra       I analog
#define EyeRf   0b00001000       // AN2 IR detect frontale destro     I analog
#define EyeLf   0b00001100       // AN3 IR detect frontale sinistro   I analog
#define EyeLl   0b00010100       // AN5 IR detect laterale destro     I analog
#define EyeRl   0b00011100       // AN7 IR detect laterale sinistro   I analog
#define FrC     0b00011000       // AN6 FotoResistenza centrale       I analog
#define GasIn 	 0b00010000		   // AN4 Gas									 I analog

										  						 
#define LedVerdeON LedVerde=0
#define LedVerdeOFF LedVerde=1
#define LedRossoON LedRosso=0
#define LedRossoOFF LedRosso=1
#define LedGialloON LedGiallo=0
#define LedGialloOFF LedGiallo=1
#define DeadCornerTime 5000	// finestra di tempo di analisi delle collisioni
#define MaxDeadCorner  10 // limite di urti su bumpers o sensori IR


//----- Macro
#define abs(x) ((x) > 0 ? (x) : -(x)) // funzione abs() valida per int, long, float...

/* Signed divide by power of two
	mantiene i bit di segno in quanto il compilatore non li propaga nel caso di shift
	di numeri negativi
*/
#define DivIntS64(x) ((x) >= 0 ? (x>>6) : ((x>>6) | 0xFC00))	// div 64 signed int 
#define DivLongS16(x)  ((x) >= 0 ? (x>>4) : ((x>>4) | 0xF0000000))	// div 16 signed long 
#define DivLongS128(x) ((x) >= 0 ? (x>>7) : ((x>>7) | 0xFE000000))	// div 128 signed long 
#define DivLongS256(x) ((x) >= 0 ? (x>>8) : ((x>>8) | 0xFF000000))	// div 256 signed long 

//----- Definizioni variabili I2C
#define  I2cEventFlag VARbits2.bit4   // la porta SSP ha comunicato un evento tramite ISR
#define  I2cBusCollFlag VARbits2.bit5 // la porta SSP ha comunicato un bus collision tramite ISR
#define  I2cBusyFlag VARbits2.bit6    // e' in corso una comunicazione I2c (N byte Tx e/o Rx)

#define	I2cMaxDev 6				// n di device I2C da controllare

unsigned char I2cDevPtr = 0;	// puntatore al device I2C in uso

const unsigned char I2cAdr[] = {0x00, 0xC0, 0x50, 0x40, 0x42, 0xE0}; //indirizzi dei device I2C

struct Nibble {
	unsigned char Rx:4;			// semibyte da usare dove bastano 16 stati
	unsigned	char Tx:4;			 
	};
		
struct I2cS {
	struct Nibble Flag;			// Rx 0=idle, 1=1 Byte da ricevere, 2=2 Byte da ricevere
										// Tx 0=idle, n=n Byte da trasmettere(con n!=0)
	unsigned char RxBuff[2];   // buffer Byte(s) da ricevere
	unsigned char TxBuff[4];	// buffer Byte(s) da trasmettere
	};

struct I2cS I2c[I2cMaxDev];
struct Nibble Ptr;				// puntatore per buffer byte di scambio I2c (Rx e TX)
char I2cStat;						// variabile di stato sequenza I2c

	// codici stato eseguito

	enum I2cBusStates{START, WRITE, ADRR, READ, ACK, NACK, RSTART, STOP, FINE};
		
// Indirizzi puntatori devices I2c
#define ExtEepromPtr 0
#define CmpPtr 1
#define AdExpanderPtr 2
#define LcdPtr 3
#define KeypadPtr 4
#define SonarPtr 5

 
 
//----- Definizioni variabili LCD
struct {
  unsigned LcdBL:1;	// Back Light-> 1 = ON
  unsigned LcdRS:1;	// Instruction/Data register selection-> 0=instruction 1=data
  unsigned LcdRW:1;	// Read/Write selection-> 0=reg.write, 1=reg.read
  unsigned LcdNA:1;	// Non Assegnato
  unsigned LcdData:4;// Data I/O  lines (4 bit mode)
  } LcdBits;
 
#define LcdBL LcdBits.LcdBL
#define LcdRS LcdBits.LcdRS
#define LcdRW LcdBits.LcdRW
#define LcdNA LcdBits.LcdNA
// LcdEN = Port D7
#define LcdData LcdBits.LcdData

#define LcdPutCharFlag VARbits2.bit7 // Flag abilitazione routine LcdPutByte
#define Lcd4_8BitFlag VARbits3.bit0  // Flag scelta  modalita' 4 o 8 bit
#define InitFlag VARbits3.bit1	 	 // Avvia la  routine di inizializzazione
#define LcdCycle 100						 // Tempo di visualizzazione delle variabili(mS)

unsigned char LcdByteStatus;// Contatore stati per la scrittura sull'LCD
unsigned char LcdInitStatus;// Contatore stati per la routine di inizializzazione
unsigned char LcdStatus;	 // Contatore stati per visualizzazione normale
unsigned char DisplayStatus;// Contatore stati per diversi modi di visualizzazione
char LcdChar;		 			 // Il carattere (dato o comando) da inviare all'LCD
unsigned char LcdStringPtr; // Puntatore al carattere della stringa in stampa
unsigned char LcdString[34]="----------------\r----------------";// 32 caratteri + CR + EOL
char  *LcdBtoa="-999";    	 // stringa per la conversione da char a ascii
int TimerLcd;         		 // per le temporizzazioni dell'LCD

/* tabella stringhe su LCD
   0 = Eye
   1 = Compass
   2 = GasLight
   3 = Threshold
   4 = PID
*/

#define MaxLcdStatusItem 5  // numero max item da visualizzare

rom const char LcdTable1[][17] = {  "                ",
                                    "ELl ELf. ERf ERl",
                                    "Compass. Bearing",
                                    "FrL FrC. FrR Gas",
                                    "L1On L2On- GasOn",
                                    " kP  kI-  kD  En" };
rom const char LcdTable2[][17] = {  "                ",
                                    "ELl ELf .ERf ERl",
                                    "Compass .Bearing",
                                    "FrL FrC .FrR Gas",
                                    "L1On L2On -GasOn",
                                    " kP  kI - kD  En",
                                    "                ",
                                    "ELl ELf .ERf ERl",
                                    "Compass .Bearing",
                                    "FrL FrC .FrR Gas",
                                    "     L2On  GasOn",
                                    "     kI   kD  En",
                                    "                ",
                                    "ELl ELf .ERf ERl",
                                    "Compass .Bearing",
                                    "FrL FrC .FrR Gas",
                                    "L1On       GasOn",
                                    " kP       kD  En",
                                    "                ",
                                    "ELl ELf .ERf ERl",
                                    "Compass .Bearing",
                                    "FrL FrC .FrR Gas",
                                    "L1On L2On  GasOn",
                                    " kP  kI       En",
                                    "                ",
                                    "ELl ELf .ERf ERl",
                                    "Compass .Bearing",
                                    "FrL FrC .FrR Gas",
                                    "L1On L2On       ",
                                    " kP  kI   kD    " };
                                    
unsigned char LcdVar[MaxLcdStatusItem+1][5];// variabili da mostrare sul display

//----- Definizioni variabili Keypad
#define  FlagKbdIntr VARbits3.bit3	      // Flag abilitazione routine Keypad
#define  FlagKbdStartRead VARbits3.bit5	// Flag avvio lettura Keypad
#define  FlagKbdStartI2c VARbits4.bit0    // Flag per attesa lettura I2c

unsigned char CursorStatus;			// Posizione cursore nei menu con modifica
unsigned char TimerKeypad;          // Timer per debounce tasti
unsigned char EepromStatus;         // Contatore stati per scrittura su EEPROM
  
#define  UserInterfaceFlag VARbits3.bit6	// Flag di esecuzione routine interfaccia utente
#define  FlagMenuMod VARbits4.bit1	      // Flag Menu Modifica
#define  EepromFlag VARbits3.bit7      	// Flag di esecuzione routine EEPROM


//----- Definizioni variabili bussola
#define CmpBearing LcdVar[2][0] // angolo letto dalla bussola
int TimerCmp = 0;               // timer lettura bussola
#define CmpCycle 100;           // periodo lettura bussola in ms
#define  FlagCmp VARbits4.bit2  // Flag esecuzione routine bussola
#define  FlagCmpCal VARbits4.bit3  // Flag esecuzione routine calibrazione bussola
#define  FlagCmpReg15 VARbits4.bit4  // Flag per scrittura in registro di calibrazione


/* Fine definizioni+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/* Prototipi funzioni *****************************************************************/

void CmpCal (void);

void CmpRead (void);

void Display (void);

void WriteEeprom(unsigned char Data, unsigned char Adr);

unsigned char ReadEeprom (unsigned char Adr);

void UserInterface(void);

void TimeKeeping (void);

void SequenzaStart (void);

void LcdPutString(void);

void LcdPutStringInitC2A(unsigned char Cursor,unsigned char Val);

void LcdPutStringInitRom(unsigned char Cursor, const char rom *LcdStr);

void LcdPutChar (char LChar, unsigned char LRS, unsigned char L4_8Bit);

unsigned char LcdByteSet (void);

void Lcd4Bit (void);

void Lcd8Bit (void);

void LcdInit (void);

void I2cHighService (void);

void I2cLowService (void);

void Target (char Target);

void Beep (void);

void SegueLuce (void);

void Light (void);

void Sound (void);

void Gas (void);

void Eye (void);

void Settings(void);

void ReadAD (void);

void MotSpeed (void);

void Turn(long Gradi);

void Walk(long Space);
 
void Space2Run (void);

void Stop(void);

void Path (void);

void Bumpers(void);

void InterruptHandlerHigh (void);

void InterruptHandlerLow (void);

/* Fine prototipi funzioni ***********************************************************/

/*Settings*****************************************************************************
                                                            setting di porte e registri
*/
void Settings(void)
{
	
//-------Port A

TRISA   =   0b11111111;   // tutti input (anche quelli non usati)

//-------Port B

TRISB   =   0b00111111;   // tutti input
INTCON2bits.RBPU;			  // PORTB Pull-up enabled


//-------Port C

TRISC = 0b10011001;      
/*
RX
TX
Buzzer
SDA
SCL
PWMR
PWML
don't care
*/


//-------Port D

TRISD = 0b00001110; 
/*
bit 0, 4, 5, 6, 7 come output
gli altri come input


//-------Port E

TRISE = 0b00000111;
/*
don't care
don't care
don't care
PORTD in I/O mode
don't care
all PORTE as input
*/


//-------A/D converter
ADCON1bits.VCFG1=0;	// VCFG1:VCFG0 Voltage Reference Configuration bit (VREF- source) 0=Vss
ADCON1bits.VCFG0=0;  //             Voltage Reference Configuration bit (VREF+ source) 0=Vdd
ADCON1bits.PCFG3=0;  // PCFG3:PCFG0 A/D Port Configuration Control bits
ADCON1bits.PCFG2=1;  //             From AN0 to AN7 Analog ports.
ADCON1bits.PCFG1=1;  //             From A8 to AN12 Digital ports.
ADCON1bits.PCFG0=1;
ADCON2bits.ADFM=0;	// la lettura e' giustificata a sinistra, gli 8 bit piu' significativi sono
            			// in ADRESH gli ultimi 2 bit (che possono non essere utilizzati se sono  
            			// sufficienti 256 livelli) sono i bit 7 e 6 di ADRESL
ADCON2bits.ADCS2=1;  // ADCS2:ADCS0 A/D Conversion Clock Select bits
ADCON2bits.ADCS1=1;
ADCON2bits.ADCS0=1;
ADCON2bits.ACQT2=0;  // ACQT2:ACQT0 A/D Acquisition Time
ADCON2bits.ACQT1=0;  // ACQT2:ACQT0 = 000 Manual acquisition Time
ADCON2bits.ACQT0=0;  // compatible with device without acquisition time.
ADCON0bits.ADON=1;	// modulo A/D acceso



//-------Interrupts

RCONbits.IPEN=1; 			//	abilitata priorita' interrupt, non compatibile con 16fxx
INTCONbits.GIEH=0;		// disabilitati tutti gli interrupt high
INTCONbits.GIEL=0;		// disabilitati tutti gli interrupt low	


//-------Interrupt su PortB

INTCONbits.INT0IE=1;		// INT0 abilitato, Keypad 
INTCON3bits.INT1IE=1;	//	INT1 abilitato, PortB1 encoder ticks
INTCON3bits.INT2IE=1;	//	INT2 abilitato, PortB2 encoder ticks

								// INT0 sempre ad alta priorita'
INTCON3bits.INT1IP=0;	// INT1 bassa priorita'
INTCON3bits.INT2IP=0;	// INT2 bassa priorita'

INTCON2bits.INTEDG0=0;	// INT2 sul fronte di discesa 
INTCON2bits.INTEDG1=1;	// INT1 sul fronte di salita (se verso errato cambiare in 0)
INTCON2bits.INTEDG2=1;	// INT2 sul fronte di salita 

INTCONbits.RBIE=0;		// interrupt on change PortB disabled


//-------Timer0

TMR0L -= 156;
/*interrupt ogni 1mSec 
periodo = CLKOUT * prescaler * TMR0
  1mSec =~ 1/(20.000.000/4) * 32 * 156 @ 20MHz
  1mSec =~ 1/(40.000.000/4) * 64 * 156 @ 40MHz  
*/

#ifdef PLL // 40MhZ
	T0CON  = 0b11000101;
#else
	T0CON  = 0b11000100;
#endif

/*
TMR0ON	TMR0 acceso
T08BIT	Timer a 8 bit
T0CS		Timer on internal clock
TOSE		increment on low-to-high transition	
PSA		Prescaler assegnato
prescaler 1:32 per TMR0 = 1ms con clock a 20MHz
prescaler 1:64 per TMR0 = 1ms con clock a 40MHz
*/

INTCONbits.TMR0IE=1;			// interrupt on TMR0 overflow enabled
INTCON2bits.TMR0IP=1;		// TMR0 interrupt high priority


//-------USART

#ifdef PLL // 40MhZ
	SPBRG=129;              //19200 @ 40MHz
#else
	SPBRG=64;               //19200 @ 20MHz
#endif

TXSTAbits.BRGH=1;              	//High baud rate 
TXSTAbits.SYNC=0;						//asynchronous
RCSTAbits.SPEN=1;						//enable serial port pins
RCSTAbits.CREN=1;						//enable reception
RCSTAbits.SREN=0;						//no effect
PIE1bits.TXIE=0;						//disable tx interrupts
PIE1bits.RCIE=0;						//disable rx interrupts
TXSTAbits.TX9=0;						//8-bit transmission
RCSTAbits.RX9=0;						//8-bit reception
TXSTAbits.TXEN=0;						//reset transmitter
TXSTAbits.TXEN=1;						//enable the transmitter


//-------PWM

CCP1CON=0b00001100;  //ccpxm3:ccpxm0 11xx=pwm mode (compatible mode)
CCP2CON =0b00001100; //ccpxm3:ccpxm0 11xx=pwm mode
T2CON = 0X01;        //STOP TIMER2 registers to POR state + prescaler = 4
PR2 = 0x7F;          // PWM freq = 19KHz @ 40 MHz

T2CONbits.TMR2ON = 1;// Turn on PWM

//-------I2C

OpenI2C(MASTER, SLEW_OFF);	//Master I2C, slew rate off per clock=100KHz
SSPADD=99;						//100KHz baud clock @ 40MHz
									//SSPAD=((Fosc/BitRate)/4)-1=(400/4)-1=99
PIE1bits.SSPIE=1;				// SSP (I2C events) int enabled
PIE2bits.BCLIE=1;				// BUS COLLISION int enabled
IPR1bits.SSPIP=0;				// SSP int = low priority
IPR2bits.BCLIP=0;				// BUS COLLISION int = low priority


//-------Interrupts non usati

PIE1bits.TMR2IE=0;	   	// interrupt on TMR2 overflow disabled
PIE1bits.TMR1IE=0;   		// interrupt on TMR1 overflow disabled
PIE2bits.TMR3IE=0;			// interrupt on TMR3 overflow disabled
PIE1bits.PSPIE=0;				// PSP int disabled
PIE1bits.ADIE=0;				// AD int disabled
PIE1bits.RCIE=0;				// USART RX int disabled
PIE1bits.TXIE=0;				// USART TX int disabled
PIE1bits.CCP1IE=0;		   // CCP1 int disabled
PIE2bits.CCP2IE=0;		   // CCP2 int disabled
PIE2bits.EEIE=0;				// EEPROM int disabled
PIE2bits.LVDIE=0;				// LOW VOLTAGE int disabled


//-------Interrupts
		//INTCON3bits.INT1IF			// interrupt flag encoder SX
		//INTCON3bits.INT2IF			// interrupt flag encoder DX	
		//INTCONbits.INT0IF			// interrupt on INT0 
		//PIR1bits.TMR2IF	  		 	// interrupt on TMR2 overflow
		//PIR1bits.TMR1IF  	 		// interrupt on TMR1 overflow
		//PIR1bits.PSPIF				// PSP int
		//PIR1bits.ADIF				// AD int
		//PIR1bits.RCIF				// USART RX int 
		//PIR1bits.TXIF				// USART TX int 
		//PIR1bits.SSPIF				// SSP int 
		//PIR1bits.CCP1IF			   // CCP1 int 
		//PIR2bits.CCP2IF			   // CCP2 int 
		//PIR2bits.EEIF				// EEPROM int 
		//PIR2bits.BCLIF				// BUS COLLISION int 
		//PIR2bits.LVDIF				// LOW VOLTAGE int 
		//PIR2bits.TMR3IF				// interrupt on TMR3 overflow 

}
/*fine settings************************************************************************/

/* ///////////////////////////////////////////////////////////////////////////////////
** Continua descrizione:
**  
** Processore PIC 18F452, controller board Mark III
**
**   e' previsto l'utilizzo dei seguenti sensori:
**      4 sensori di prossimita' IR sharp GP2D120        4 I   (EyeRf,EyeLf,EyeRl,EyeLl)
**																			      AN2-AN3-AN5-AN7
**      2 sensori d'urto a microswitch (bumper)          2 I   (BumperR, L)      D2-D3
**      2 shaft encoder in quadratura                    4 I   (EncoderRdir,L)   B1-B2
**                                                             (EncoderRpulse,L) B3-B4
** 	  3 Sensori di luce                                3 I   (FrL,FrC,FrR)
**																					AN0-AN1-AN6
** 	  Sensore di suono                                 1 I   (Sound)				D1
** 	  Sensore di gas                                   1 I   (Gas)					AN4
**
**   e dei seguenti attuatori:
**    H-bridge motor driver                              3 O   (PWMmotR, L)      C1-C2
**                                                             (MotEnable)       D0
**      Buzzer                                           1 O                     C5 
**   delle porte Rx-Tx                                   2 I/O
**
**   oltre ai seguenti dispositivi I2c (2 I/O SCL-SDA)
**      Ultrasonic range finder Devantech SRF08         Adr =	0xE0 (224)       
**      256Kbit I2C EEprom                              Adr =	0x00 (000)
**      I2C 8bit I/O expander PCF8574                   Adr =	0x40 (064)
**      I2C 8bit I/O expander PCF8574                   Adr =	0x42 (066)
**      Analog input expander                           Adr =	0x50 (080)
**      Digital compass Devantech CMPS03                Adr =	0xC0 (192)
**
**
** A0(AN0) FrR                      I   analog
** A1(AN1) FrL                      I   analog
** A2(AN2) EyeRf                    I   analog
** A3(AN3) EyeLf                    I   analog
** A4      ------------------
** A5(AN4) Gas								I   analog
** B0      I/O expander interrupt	I
** B1      EncoderRpulse            I
** B2      EncoderLpulse            I
** B3      ------------------
** B4      EncoderRdir              I   
** B5      EncoderLdir              I
** B6      PGC (ICD2)				 
** B7      PGD (ICD2)	
** C0      ------------------
** C1      PWMmotR                  O   PWM   
** C2      PWMmotL                  O   PWM
** C3      SCL                      I
** C4      SDA                      I
** C5      Buzzer                   O   
** C6      TX                       O
** C7      RX                       I
** D0      MotEnable                O
** D1      Sound							I
** D2      BumperR                  I
** D3      BumperL                  I
** D4      LedRosso----------			O	temp
** D5      LedVerde----------			O	temp
** D6      LedGiallo---------			O	temp
** D7      LCD enable					O
** E0(AN5) EyeRl                    I   analog      
** E1(AN6) FrC                      I   analog
** E2(AN7) EyeLl                    I   analog

----------------------------------------------------------------------------------------
lettura sensori  IR Sharp GP2D120
distanza dal paraurti

distanza    cm:25 24 23 22 21 20 19 18 17 16 15 14 13 12 11 10 09 08 07 06 05 04 03 02
lettura AD hex:10 11 13 14 16 18 1a 1e 20 24 26 2b 30 33 36 3b 3f 46 4e 5a 66 77 8d a2   

sotto i 2cm la  tensione ricomincia a  scendere.
----------------------------------------------------------------------------------------

----------------------------------------------------------------------------------------
Descrizione delle metodologie generali di programmazione

Il programma lavora in una versione semplificata di "Real Time Operating System",
sono state seguite quindi le seguenti regole per lo sviluppo di un
"Sistema Operativo Multitasking Cooperativo":

-Ogni routine  collabora con tutte le altre, ed in particolare con il main, 
   per il buon funzionamento del sistema. 

-Ogni routine occupa il sistema per il minor tempo possibile 

-Non ci sono loop lunghi o ritardi a SW 

-Il main si preoccupa di schedulare i vari task, le routine si chiamano raramente tra loro 

-Lo scambio di informazioni tra task avviene tramite flag e variabili globali 

-Le temporizzazioni sono realizzate usando come "Real Time Clock" l'interrupt overflow 
   del TIMER0 che e' il vero "hearth-beat" del sistema 

-L'esecuzione dei vari task e' condizionata dai relativi semafori, il main o un altro task 
   possono disabilitare una certa funzione 

----------------------------------------------------------------------------------------
Descrizione delle procedure di movimento:

Il cuore di tutto e' il timer (TMR0) che ogni 1mSec genera un interrupt.
Nella ISR vengono incrementati i timer (a 8, 16, 24 bit secondo la durata necessaria)
che servono per le temporizzazioni delle varie routine.

Anche il computo dello spazio percorso tramite encoder ottici in quadratura, funziona ad 
interrupt,ognuno dei due encoder genera il proprio interrupt che, analizzato e confrontato 
con il flag di direzione, incrementa o decrementa le variabili di conteggio dello spazio.

Nella stessa ISR viene calcolata, ogni 50mSec, la velocita' di avanzamento attuale,
dividendo lo spazio percorso per il tempo.

Il main e' semplicemente uno scheduler che lancia le diverse routine con le giuste priorita'.

Una di queste e' la MotSpeed che lavora in background e, ogni 50 mSec (dopo la misura della
velocita'), imposta il PWM (calcolato tramite un algoritmo PID) per fare in modo che la 
velocita' attuale corrisponda con quella desiderata, .

A questo punto abbiamo un sistema a loop chiuso nel quale sappiamo la velocita', lo spazio
percorso ed il tempo trascorso.

Anche la routine Path lavora in background e, se abilitata, controlla la sequenza di 
operazioni da effetuare per seguire il percorso impostato da qualche altra routine che
decide il comportamento del bot.
La routine che decida di eseguire un percorso diverso dal normale, deve:
- impostare il flag di attivazione della routine Path
- scrivere nell'array PathSeq la sequenza di codici della manovra da effettuare 
   (0=fine sequenza)
- azzerare il puntatore della sequenza
Path comincera' allora a scandire l'array chiamando le routine in funzione del codice 
impostato (cammina X cm, ruota Y gradi, cammina W mm/sec). 
La routine chiamata (Walk, Turn) calcola lo spazio che deve percorrere ogni ruota prima 
di ripassare il comando a Path.
A questo punto si e' attivata Space2Run (schedulata da main) che si disattiva, ripassando
il controllo a Path, solo quando le ruote si sono mosse dello spazio desiderato.
Path controlla la sequenza successiva ripetendo le operazioni o restituendo il controllo
quando il codice e' = 0.

Il Main e' sempre attivo e, visto che le operazioni di movimento sono molto lente rispetto
ai tempi di esecuzione del PIC, per la maggior parte del tempo sara' occupato dall'interrupt
degli encoder: 
a 36cm/s le ruote fanno 2 giri/s, l'encoder fa 40 giri/s, con 40 finestrelle genera
1600 impulsi/s (solo fronte di salita) -> 3200 tick/s da entrambi gli encoder
-> un impulso ogni 312,5us. 
Con il clock a 40MHz (100ns a istruzione) -> 3125 istruzioni tra un interrupt e l'altro.


----------------------------------------------------------------------------------------
Descrizione delle procedure di comunicazione I2C

Indirizzi devices I2c
// #	device	descrizione		|fix	 | |free| |R/W|	|adr|
//	0	24LC256	32KB eeprom		0 0 0 0	0 0 0	  x		0x00
//	1	CMPS03	compass			1 1 0 0  0 0 0	  x		0xC0
//	2	MAX127	A/D expander	0 1 0 1	0 0 0	  x		0x50
//	3	PCF8574	I/O expander	0 1 0 0	0 0 0	  x   	0x40
//	4	PCF8574	I/O expander	0 1 0 0	0 0 1	  x		0x42
//	5	SRF08		Devantech US	1 1 1 0	0 0 0	  x		0xE0

Analizzando i sorgenti delle librerie I2c del C18 Microchip si nota che alcune contengono 
dei waiting loop in attesa dell'esecuzione dell'operazione da parte della SSP e non sono
quindi compatibili con il sistema pseudo-RTOS

ReadI2C		wait until byte received
getcI2C		idem
getsI2C		perform getcI2C() for 'length' number of bytes
      		check that receive sequence is over
				wait until ACK sequence is over
IdleI2C		Test and wait until I2C module is idle
WriteI2C		wait until write cycle is complete
putI2C		idem
putsI2C		transmit data until null character
				test for idle condition
				wait until ninth clock pulse received

altre invece possono essere usate 
AckI2C		ok
CloseI2C		ok
DataRdyI2C	ok
NotAckI2C	ok
OpenI2C		ok
RestartI2C	ok
StopI2C		ok
StartI2C		ok

Tutte le funzioni relative alla eeprom contengono molti wait on idle

Tutte le funzioni relative agli LCD contengono molti delay

Le procedure non utilizzabili sono quindi state riscritte in modo compatibile  usando gli
interrupt generati dall'SSP, con flag di stato e variabili globali di appoggio.

Un'ottima traccia di partenza e' stata l'application note AN736 di Microchip che descrive
in modo molto dettagliato questo tipo di tecnica.

La comunicazione tramite I2c avviene a diversi livelli.
		
-Strato applicativo: ogni device I2c ha le sue caratteristiche, 
 	la routine relativa al device	ha a disposizione 4byte in TX e 2byte in RX per ogni turno.
	Ad esempio: verso la EEPROM si trasmetteranno due byte per l'indirizzo della locazione 
	di memoria e si riceveranno o trasmetteranno uno o due byte (variabile char o int).
	E' questo il device che richiede piu' byte in trasmissione e determina quindi la dimensione 
	del buffer.
	La funzione I2cDeviceXX prepara i byte da scambiare e alza i flag relativi alle azioni 
	da svolgere:
	I2c[I2cDevPtr].Flag.Rx  
	I2c[I2cDevPtr].Flag.Tx .
		
		
-Strato I2cHighLevel: 
	E' eseguito se il deviceXX ha qualcosa da scambiare
	I2c[I2cDevPtr].Flag.Rx !=0 || I2c[I2cDevPtr].Flag.Tx!=0
	e se la routine a livello piu' basso e' disponibile:
	I2cBusyFlag = 0
			
	scambia i byte con il device attraverso la routine a livello piu' basso alza il I2cBusyFlag 
			
			
-Strato I2cLowLevel:
	E' eseguito se la ISR ha alzato I2cEventFlag
	
	START->ADRW->								se ha qualcosa da trasmettere, altrimenti Rx
	WRITE byte[0]->...WRITE byte[n]->
	RSTART->ADRR->
 	READ byte[0]->ACK...READ byte[n]NACK->
 	STOP
	FINE


Nel  ciclo main sono controllati i flag a partire da quelli a livello piu' basso a salire.
L'intero ciclo e' ripetuto  per ogni device I2c scandendo il buffer 

1-Evento I2c ? YES -> esegui I2cLowService -> EXIT
	|
	V
	NO 
2-I2cBusyFlag ? YES -> EXIT
	|
	V
	NO
3-C'e' qualche byte da trasmettere e/o da ricevere ? -> YES -> I2cHighService -> EXIT
	|
	V
	NO
4-Ci sono ancora record da controllare nel buffer I2c ? -> YES -> Incrementa contatore I2cDevPtr  -> EXIT
	|
	V
	NO
5-Azzera contatore I2cDevPtr -> EXIT

descrizione
1-L'evento I2c e' comunicato da SSP tramite ISR
	I2cLowService esegue sequenza I2c per scambiare singolo byte
		azzera flag evento o collisione I2c
		scambiato l'intero byte azzera flag I2cBusyFlag per abilitare  la routinee high
		
2-I2cBusyFlag e' alzato da I2cHighService e azzerato da I2cLowService

3-La routine relativa al singolo  device I2c ha riempito il buffer ed alzato i flag
	I2cHighService inizializza sequenza I2cLowService, questa scambia ogni singolo byte
		ad ogni byte scambiato e' decrementato il contatore relativo
		quando contatore TX e' a zero si inizia la ricezione
		quando entrambi i contatori sono a zero si passa al device successivo
	solo se entrambi i contatori sono a zero la routine del singolo device, quella a 
		livello piu' alto, pu cominciare a scambiare altri byte
		
4-E' stato scandito tutto il buffer, al prossimo giro si ricomincia dal primo

5-Si prepara a controllare il device successivo al prossimo giro


////////////////////////////////////////////////////////////////////////////////////// */

/* Version History /////////////////////////////////////////////////////////////////////


0.1 - inizio lavori 19 luglio 2003

0.2 - 16 agosto 2003
prime prove di movimento, prime implementazioni dell'algoritmo PID

0.3 - 14 aprile 2004
prima versione dimostrabile e pubblicata sul sito. Implementate le funzioni di movimento di base.
Il PID funziona bene, riesco a far muovere il bot ad una precisa velocita' e farlo girare dell'angolo
che voglio come dimostrato col programma"quadrato.c" e col video pubblicato sul sito.
Implementato riconoscimento ostacoli con bumper ed algoritmo per uscire dagli angoli morti

0.4 - inizio 30 novembre 2004
Cambiato boot loader con versione di Peter Huemer, Shane Tolmie e Pic Bootloader+ di Herman Aartsen.
Questo ha un ritardo di soli 0.2 secondi all'accensione che non provocano movimento inutile del bot
all'avvio (5 sec nella versione inizialmente caricata nel MarkIII).
Ho modificato la versione originale per fare in modo che i motori non girassero durante la programmazione:
Motenable = 0
Implementazione riconoscimento ostacoli tramite sensori infrarossi Sharp GPxxx.
Cambiati sensori originali a lungo raggio perche' iniziano la misura da 20cm. Vanno meglio i GP2D120 che
partono da 4cm. Piu' che sufficiente la distanza massima di 40cm.
Implementate le funzioni di lettura A/D.

0.41 - inizio 17 gennaio 2005
Dopo il cambio di motori e alcune prove sul campo a Pisa, ho cambiato il rapporto di riduzione degli
ingranaggi per aumentare la velocita' di esplorazione del campo. Da 269:1 sono passato a 101:1,
passando da una velocita' teorica massima di 260cm/s ad una di 690cm/s, i motori hanno ancora sufficiente
potenza e questo mi dovrebbe permettere un buon aumento di velocita' con ancora un alto margine di controllo.
28-01-2005
Ho tolto i sensori "LineX" sul paraurti, inutili per un robot explorer e ho aggiunto altri 2 sensori IR Sharp
GP2D120 per rivelare ostacoli frontali usando le porte A/D precedentemente assegnate ai sensori di linea.
Ho aggiunto anche tre fotoresistenze al CdS per rivelare le sorgenti di luce

0.42 - inizio 21 marzo 2005
modificato algoritmo per uscire dagli angoli morti per adattarlo ad un diverso numero e tipo di sensori

0.5 - 28 marzo 2005
Trovati falsi allarmi dovuti ai sensori IR che restituiscono impulsi quando sono investiti dalla sorgente di luce.
Inserito debounce su rivelazione luce per diminuire falsi allarmi.
Quando comincia a rivelare la luce diminuisce la velocita', altrimenti i tempi di intervento sono troppo ristretti
e si rischia di perdere la sorgente o di sbatterci contro

0.6 - 1 maggio 2005
Aggiunto circuito per la rivelazione del tono a 4KHz e relativa routine di gestione. Le logiche di rivelazione,
debounce e segnalazione sono simili a quelle usate per la rivelazione delle sorgenti luminose
   
1.0 - 7 maggio 2005
Con questa versione Dino ha partecipato alla prima gara all'ITIS Von Neumann con un discreto piazzamento. 
Modificato tra la prima e la seconda manche per disabilitare la ricezione del suono mentre e' in pausa dopo aver
rivelato la luce, altrimenti c'e' il rischio che entri in un loop suono-luce se rivela il tono insieme alla luce.
Tarate soglie di luce per il campo di gara. Aumentato delay dopo "TargetOK".
 
1.1 - 10 maggio 2005
Disabilitato sensore suono anche durante beep per evitare che riveli il suo tono come segnale a 4KHz
Regolati ritardi, soglie e velocita' min e MAX
   
Ha partecipato anche alla gara al Pacinotti, problema sull'HW del suono nella seconda manche, risolto.
   
2.0 - 1 giugno 2005
Installato bootloader "tiny boot loader" adatto per Pic18, occupa solo 100byte ricompilabile per la frequenza
di clock voluta.
Iniziato porting su PIC18F452 con compilatore Microchip C18 e clock a 40MHz

18_1.0 - 8 giugno 2005
Porting finito in imitazione 16f877 anche con clock a 40MHz
   
18_2.0  - 13 giugno 2005
Modificata gestione interrupt sfruttando le caratteristiche del 18f452
TMR0 e' gestito come interrupt ad alta priorita'.
Encoder gestiti come interrupt a bassa priorita' su INT1 e INT2, solo il fronte di salita genera interrupt,
sono quindi dimezzati gli interrupt generati rispetto alla versione precedente.
Non cambia il cablaggio rispetto alle versioni precedenti, sono solo invertiti EncoderXPulse con EncoderXDir.
   
18_2.1 - 17 luglio 2005
Iniziato sviluppo comunicazione I2C
Nelle procedure di movimento sono state ricalcolate le costanti ragionando in base 2, in  questo modo si sono
sostituite le divisioni con gli shift, risparmiando centinaia di microsecondi a procedura.

18_2.2 - 20 agosto 2005
Sviluppate procedure comunicazione I2C
Sviluppato HW e SW per I/O expander per pilotare LCD e tastiera

18_2.3 - 3 settembre 2005
Sviluppate procedure per la scrittura su display LCD

18_2.4 - 5 settembre 2005
Cambiata procedura inizializzazione, il ciclo idle prima della partenza e' gia 
nell'RTOS.
Iniziato sviluppo routine rivelazione gas			

18_2.41 - 12 settembre 2005
Terminate procedure rivelazione gas
Lcd enable collegato a porta RD7 e non piu' a I/O expander per problema temporizzazioni:
lo strobe all'LCD (En=1, En=0) deve essere eseguito quando il dato e' stabile sul bus,
il PCF8574 garantisce la stabilita' dei segnali sulla porta d'uscita solo dopo 4uS
dall'ultimo colpo di clock dopo l'ACK.
Capitava spesso che il dato sull'LCD non fosse corretto, facendolo impazzire. 

18_2.42 - 1 novembre  2005
Ottimizzate procedure LCD.
Cambiata Interrupt Service Routine ad alta priorita':
   Time keeping esterno all'ISR
   Interrupt dell'I/O  expander 2 (Keypad) su INT0
Sviluppate procedure di base per keypad.
Tolto bootloader. programmazione e debug con ICD2

18_2.43 - 16 novembre  2005
Ottimizzata procedura interpretazione e debounce tasti.

18_2.44 - 17 novembre 2005
Suddiviso codice in piu' file per facilitarne la scrittura

18_2.45 - 3 dicembre 2005
Ottimizzato main raggruppando le user interfaces 
   (LCD, keypad, ecc...) sotto un unico semaforo.
Finita gestione display e tastiera con menu visualizzazione delle variabili
   e modifica e salvataggio in EEPROM delle costanti
   
18_2.50 - 01 gennaio 2006
Porting su 18F4525
   il grosso del cambiamento e' sulla configurazione dell'A/D converter, 
   nell'utilizzo non e' cambiato niente
Finite procedure bussola
   lettura bussola (1 byte -> 0-360 = 0-255) ogni 100 ms
   calibrazione automatica bussola sui quattro punti cardinali
   
18_2.51 - 01 gennaio 2006
Inizio procedure odometria

/////////////////////////////////////////////////////////////////////////////////// */

/*fine descrizioni ////////////////////////////////////////////////////////////////////// */
	

