/*
contiene tutte le routine per la gestione della tastierina e del display LCD
*/


/*UserInterface ***********************************************************************
Raggruppa tutte le routine gestite manualmente: Keypad, LCD, EEPROM.
Dal momento che sono eseguite molto raramente sono tutte ragruppate sotto un unico
semaforo. 
Questo permette di risparmiare diversi cicli di controllo nel funzionamento normale.
*/

void UserInterface (void)
{
//EEPROM=========================================
if (EepromFlag)      // se = 0 e' disabilitata la gestione dell'eeprom
{
   if (!EECON1bits.WR)   // se e' finita la scrittura della variabile 
   {                 // scrive la successiva
      switch (EepromStatus)
	   {     
   	   case 0:
   	      WriteEeprom(Light1,EepromStatus);
   	      EepromStatus++; // variabile successiva
		   break;
		   
		   case 1:
   	      WriteEeprom(Light2,EepromStatus);
   	      EepromStatus++; // variabile successiva
		   break;
		   
		   case 2:
   	      WriteEeprom(GasOn,EepromStatus);
   	      EepromStatus++; // variabile successiva
		   break;
		   
		   case 3:
   	      WriteEeprom(kp,EepromStatus);
   	      EepromStatus++; // variabile successiva
		   break;
		   
		   case 4:
   	      WriteEeprom(ki,EepromStatus);
   	      EepromStatus++; // variabile successiva
		   break;
		   
		   case 5:
   	      WriteEeprom(kd,EepromStatus);
   	      EepromStatus++; // variabile successiva
		   break;
		   
		   case 6:
   	      WriteEeprom(En,EepromStatus);
   	      EepromStatus++; // variabile successiva
		   break; 
		   
		   default:
   	      EepromFlag=0;// una volta salvate tutte le variabili disabilita la routine
   	      EECON1bits.WREN=0;  // e disabilita la scrittura su EEPROM
   	      if (En)             // abilitazione motori scritta in EEPROM 
            {                   // per effettuare le misure senza far girare le ruote
               MotEnable=1; 	  // avvia motori, per debug=0, altrimenti = 1
            }
            else
            {
               MotEnable=0;
            }
		   break;
		}  // end switch
	}
}


//Init e Lcd Display=============================
if (DisplayStatus)	// se = 0 sono disabilitate tutte le routine relative al display
{
   if (InitFlag) 	
   /* Inizializza l'LCD e le routine di movimento.
   	Il flag e' settato solo prima di entrare nel main,in questo modo l'inizializzazione
   	dell'LCD avviene una sola volta.
   	Rimane in loop ciclando i led colorati fino a quando non si tocca un pulsante
   */
   
   {
		if  (BumperDx && BumperSx) // resta fermo finche' non si tocca  un baffo
		{
        	if (TimerLcd <= 0) 		// Per le temporizzazioni necessarie all'init dell'LCD
	   	{
		   	if (!LcdPutCharFlag)	// e' terminata la trasmissione di un byte
		   	{
			   	if (LcdStringPtr== 0xFF) // e' terminato l'invio della stringa
			   	{
	   				LcdInit();	   				
	   			}
	   		}
	   	}
	   }
	   else
	   {
		   if (FlagCmpCal)    // e' stata richiesta la calibrazione della bussola
		   {
   		   CmpCal();
   		}
   		else               // avvio normale
		   {
   		   SequenzaStart();// esce dallo stato init avviando le sequenze di movimento
		   }
		   if (En)            // abilitazione motori scritta in EEPROM 
         {                  // per effettuare le misure senza far girare le ruote
            MotEnable=1; 	 // avvia motori, per debug=0, altrimenti = 1
         }
		}
   }
   else	// finita la fase di init, l'LCD viene usato per mostrare i parametri voluti
   {
	  if (TimerLcd <= 0) 	// e' tempo di aggiornare il display
	  {
		 	TimerLcd=LcdCycle;// reset timer
		 	
		 	if (DisplayStatus==99)
		 	{
   		 	DisplayStatus=0;  // ha visualizzato la scritta fissa e disabilita LCD
   		}
   		else
   		{
		 	   if (DisplayStatus> MaxLcdStatusItem || DisplayStatus <= 0)
		 	   {
   		 	   LcdBL=0;          // LCD Backlight off
   		 	                     //   "0123456789012345  0123456789012345"
				   LcdPutStringInitRom(0, "----- RUN ------\r----------------");
					DisplayStatus=99;
   		   }
   		   else
   		   {
      		LcdBL=1;          // LCD Backlight on
		 	   Display(); 
		      /* Visualizza valori diversi in funzione della variabile DisplayStatus
		         
		         1: visualizza valori sensori di prossimita'
		         2: visualizza angolo bussola
		         3: visualizza valori gas e fotoresistenze
		         4: visualizza soglie Light1, Light2 e GasOn
		         5: visualizza costanti Kp, Ki, Kd, En
		         
		        99: disabilita la routine dopo la scritta fissa
		              
                  impostando un qualsiasi valore diverso da quelli elencati
						si scrive prima una stringa fissa e poi (al giro successivo)
						si disabilita la scrittura su display.
            */
            }
         }	 		 		
	  }
	}
	

//LcdPutChar ------------------------------------
   if (LcdPutCharFlag)
   {
	/*
		La routine a livello superiore ha abilitato la scrittura di un nuovo carattere
		verso l'LCD tramite LcdPutChar()
	*/
		
		if (!I2c[LcdPtr].Flag.Tx && !I2c[LcdPtr].Flag.Rx)
		/* 
			se il buffer di trasmissione I2C e' libero si pu passare un nuovo byte 
			all'I/O expander.
			Per scrivere un carattere sull'LCD occorre inviare una sequenza di byte
			tramite l'I/O expander
		*/
		{
			if (Lcd4_8BitFlag)	// modalita' 4 o 8 bit
			{
				Lcd4Bit();
			}
			else
			{
				Lcd8Bit();
			}	
		}
	}
  	

//LcdPutString ----------------------------------------------
 	if (LcdStringPtr!=0xFF)		// c'e' una stringa da scrivere sull'LCD
   {
	/*
		La routine a livello superiore ha abilitato la scrittura di una nuova stringa
		verso l'LCD tramite LcdPutStringInit()
	*/
		if (!LcdPutCharFlag)	// e' terminata la trasmissione del byte  precedente
		{
			LcdPutString();
		}
	}

}//Init e Lcd Display  
//===============================================

//Keypad-----------------------------------------
if (FlagKbdIntr)                    // E' arrivato un'interrupt dall'I/O expander
{	   	                           // al quale e' collegata la tastiera
	if (!FlagKbdStartRead)           // non e' in corso una lettura
	{
  	    FlagKbdStartRead=1;          // avvia una lettura della tastiera
	    TimerKeypad=0;               // avvia timer debounce
	    FlagKbdStartI2c=0;           // per attesa lettura I2c
 	}  
 	else
	{   	
	   if (TimerKeypad >= 20)         // debounce, aspetta 20ms 
	   {
   	   if (!FlagKbdStartI2c)       // non e' in corso una lettura I2c
	      {
	         I2c[KeypadPtr].Flag.Rx=1;// avvia sequenza lettura da I/O expander 2
	         FlagKbdStartI2c=1;
	      }
	      else
		   {
   		   if (!I2c[KeypadPtr].Flag.Rx)  // il contatore di ricezione e' vuoto
     	      {                             // e' finita la lettura I2c
	   	      FlagKbdStartRead=0;        // e' finito il ciclo debounce, 
                                          // rimane in attesa del prossimo evento
		         FlagKbdIntr=0;             // disabilita la routine, sara' riattivata dal 
                                          // prossimo interrupt, resetta anche eventuali
                                          // interrupt arrivati nel frattempo
                                          
               switch (I2c[KeypadPtr].RxBuff[0])   // tasto premuto
	            {     
   	            case 0b11111110:                 // ROSSO
      	            if (FlagMenuMod)
      			      {  // se in modalita' modifica incrementa variabile
         			      LcdVar[DisplayStatus][CursorStatus]++; 
         			   }
         			   else
         			   {  // altrimenti passa a menu successivo
            			   DisplayStatus++;
            			   LcdStatus=0;// ripassa per il ciclo clear del display
            			}
		 				break;
		 				
		 				case 0b11111101:                 // VERDE
      	            if (FlagMenuMod)
      			      {  // se in modalita' modifica decrementa variabile
         			      LcdVar[DisplayStatus][CursorStatus]--;
         			   }
         			   else
         			   {  // altrimenti passa a menu precedente
            			   DisplayStatus--; 
            			   LcdStatus=0;// ripassa per il ciclo clear del display 
            			   if (DisplayStatus >=  254)
            			   {     // indietro dal menu 0 ricomincia dall'ultimo menu
               			   DisplayStatus  = MaxLcdStatusItem;
               			}
               			else
               			{     
            			      if (DisplayStatus < 1)  
            		         {  // arrivato al primo menu disabilita terminale
               			      DisplayStatus = 90;
               			   }
               			}
            			}
		 				break;
		 				
		 				case 0b11111011:                 // BIANCO
   	               if (FlagMenuMod)
   	               {  // sposta il cursore dalla posizione 0 alla 3 ciclicamente
      	               CursorStatus++;       
      	               if (CursorStatus>3)   
      	               {
         	               CursorStatus=0;
         	            }
         	         }
		 				break;
		 				
		 				case 0b11110111:                 // BLU
   	               if (FlagMenuMod)
   	               {  // Toggle: Flag modifica menu e Lampeggio cursore LCD
      	               FlagMenuMod=0;
      	               EepromFlag=1;   // uscendo dalle modifiche abilita 
      	               EepromStatus=0; // la scrittura delle variabili su EEPROM   
      	            }
      	            else
      	            {  
         	            if (LcdVar[DisplayStatus][4])
         	            {  // Solo se il menu e' abilitato alla modifica
      	                  FlagMenuMod=1;
      	                  CursorStatus=0;
      	               }
      	            }   
		 				break;
		 				
		 			   case 0b11110000:     // tutti i tasti insieme (solo in fase di init)
   	               FlagCmpCal=1;     // avvia procedura calibrazione bussola
		 				break;

		 				default:       // se si premono due tasti contemporanemente non fa niente
		 				break;
		 				
		 			} // end switch			
            } 
 	      }
      }
   }
}	
if (!DisplayStatus && !FlagKbdIntr && !EepromFlag)
{
   UserInterfaceFlag=0; // se tutte le routine sono disabilitate, disabilita il flag
                        // generale per non entrare piu' in questo ciclo
                        // verra' riabilitato alla pressione di un tasto
}                        
                     
} // UserInterface
/**************************************************************************************/


/*Display *****************************************************************************
Visualizza valori sul display in funzione della variabile DisplayStatus
*/

void Display (void)
{
		switch (LcdStatus)
	   {     
   	case 0:
		 	TimerLcd=LcdCycle;							  	 // pausa all'avvio del ciclo
		 	LcdPutChar (0x01, 0, 4);					    // clear display
			LcdStatus ++;									    // passa allo stato successivo
			break;		
		
		case 1:
			LcdPutStringInitRom(0, &(LcdTable1[DisplayStatus][0]));// invia stringa all'LCD
		 	TimerLcd=LcdCycle;								 // reset timer
			LcdStatus ++;									    // passa allo stato successivo
			break;
			
		case 2:	
			LcdPutStringInitC2A(0x40,LcdVar[DisplayStatus][0]);// visualizza valore
		 	TimerLcd=LcdCycle;							// reset timer
			LcdStatus ++;									// passa allo stato successivo
			break;

		case 3:	
			LcdPutStringInitC2A(0x44,LcdVar[DisplayStatus][1]);// visualizza valore
		 	TimerLcd=LcdCycle;							// reset timer
			LcdStatus ++;									// passa allo stato successivo
			break;

		case 4:
		   if (FlagMenuMod)
		   {// se in modalita' modifica lampeggia il nome della variabile da modificare
   		   LcdPutStringInitRom(0, &(LcdTable2[DisplayStatus+6+(CursorStatus*6)][0]));
   		}
   		else
   		{
			   LcdPutStringInitRom(0, &(LcdTable2[DisplayStatus][0]));// invia stringa all'LCD
			}
		 	TimerLcd=LcdCycle;								 // reset timer
			LcdStatus ++;									    // passa allo stato successivo
			break;
			
		case 5:	
			LcdPutStringInitC2A(0x49,LcdVar[DisplayStatus][2]);// visualizza valore
		 	TimerLcd=LcdCycle;							// reset timer
			LcdStatus ++;									// passa allo stato successivo
			break;

		default:	
			LcdPutStringInitC2A(0x4D,LcdVar[DisplayStatus][3]);// visualizza valore
		 	TimerLcd=LcdCycle;							// reset timer
			LcdStatus = 1;									// ricomincia dal secondo passo
			break;
		
	   } // end switch	


} // Display
/**************************************************************************************/


/*LcdPutString ************************************************************************
Scrive sull'LCD, la stringa posta nel buffer
Un carattere "\r" all'interno della stringa significa "a capo" (Cursor = 0x40)
*/

void LcdPutString(void)
{
	if (LcdString[LcdStringPtr])
	{
		if (LcdString[LcdStringPtr] == '\r') // a capo
		{
			LcdPutChar(0x80+0x40, 0, 4);	// inizio seconda riga
			LcdStringPtr++;					// al prossimo giro scrive il carattere successivo
		}
		else
		{
			LcdPutChar(LcdString[LcdStringPtr], 1, 4);	// scrive nesimo carattere
			LcdStringPtr ++; 					// al prossimo giro scrive il carattere successivo
		}
	}
	else
	{
		LcdStringPtr=0xFF;	// ultimo passaggio, disabilita questa  routine
	}
		
} // LcdPutString
/**************************************************************************************/


/*LcdPutStringInitC2A **************************************************************
Inizializza la scrittura di una stringa di caratteri sull'LCD alla posizione "Cursor".
Converte una variabile unsigned char in tre caratteri ASCII
grazie a RI
*/

void LcdPutStringInitC2A(unsigned char Cursor,unsigned char Val)
{
	unsigned char u,d,c,tmp;

	LcdPutChar(0x80+Cursor, 0, 4);	// posizione di partenza del cursore	

	LcdStringPtr=0;		// Puntatore al carattere della stringa in stampa
								// se diverso da 0xFF avvia la stampa della stringa
	c=Val/100;
	tmp=(Val-100*c);
	d=tmp/10;
	u=tmp-10*d;

	LcdString[3]=0;						// EOL
	LcdString[2]="0123456789"[u];		// unita'
	LcdString[1]="0123456789"[d];		// decine
	LcdString[0]=" 123456789"[c];		// centinaia

} // LcdPutStringInitC2A
/**************************************************************************************/


/*LcdPutStringInitRom *****************************************************************
Inizializza la scrittura di una stringa di caratteri sull'LCD alla posizione "Cursor".
Dal momento che l'allocazione della memoria e' diversa tra memoria programma e memoria
ram, questa routine e' valida solo per la visualizazione di stringhe fisse.
*/

void LcdPutStringInitRom(unsigned char Cursor,const char rom *LcdStr)
{
	char count=0;			// contatore
	LcdStringPtr=0;		// Puntatore al carattere della stringa in stampa
								// se diverso da 0xFF avvia la stampa della stringa
															
	while (*(LcdStr+count))
	{
		LcdString[count] = *(LcdStr+count);
		count++;
	}
	LcdString[count]=0;		// nul character
	
	
	LcdPutChar(0x80+Cursor, 0, 4);	// posizione di partenza del cursore
	
} // LcdPutStringInitRom
/**************************************************************************************/


/*LcdPutChar **************************************************************************
inizializza flag e variabili per inviare un nuovo carattere verso l'LCD.		
Parametri:
LChar 	= carattere da scrivere (dato o comando)
LRS 		=1 se dato, =0 se comando
L4_8Bit 	=4 se modalita' 4 bit, 8 se modalita' 8 bit
*/

void LcdPutChar (char LChar, unsigned char LRS, unsigned char L4_8Bit)
{
	LcdChar = LChar;	// carattere (dato o comando)

	if (LRS)				// = 1 se Dato, =0 se Comando
	{
		LcdRS=1;
	}
	else
	{
		LcdRS=0;
	}

	if (L4_8Bit==4)	   
	{
		Lcd4_8BitFlag=1;	// = 1 se modalita' 4 bit, =0 se 8 bit
	}
	else
	{
		Lcd4_8BitFlag=0;
	}

	LcdRW=0;				// scrittura
	LcdPutCharFlag=1;	// abilita la scrittura del byte
	LcdByteStatus=0;	// azzera il contatore di stato della routine LcdXBit

} // LcdPutChar
/**************************************************************************************/


/*LcdByteSet***************************************************************************
ricostruisce il byte da inviare all'LCD tramite I2C dai singoli bit
*/

unsigned char LcdByteSet (void)
{
	
return ((((((((0x00|LcdData)<<1)|LcdNA)<<1)|LcdRW)<<1)|LcdRS)<<1)|LcdBL;

} // LcdByteSet
/**************************************************************************************/


/*Lcd4Bit *****************************************************************************
Invia un byte all'LCD nella modalita' a 4 bit tramite I2C
L'I/O expander 1 usato per pilotare l'LCD e' il device I2c numero 3
Il driver LCD HD44780 con il clock interno a 270KHz richiede una pausa di 37uSec
tra un carattere e l'altro. Per inviare un byte tramite l'I/O expander PCF8574 con il
clock a 100KHz ci vogliono almeno 90uSec (8 bit + ACK * 10uSec a bit), piu' che 
sufficienti quindi per le temporizzazioni richieste.
*/

void Lcd4Bit (void)
{
	switch (LcdByteStatus)
	{
		case (0):
			LcdEN=0;	
			LcdData=LcdChar >> 4;						// upper nibble	
			I2c[LcdPtr].Flag.Tx = 1;					// un byte da trasmettere
			I2c[LcdPtr].TxBuff[0] = LcdByteSet() ;	// compone il byte da inviare e lo passa
																// alle routine I2C. Il buffer e' stato
																// controllato prima e quindi e'
																// sicuramente vuoto
			LcdByteStatus ++;								// passa allo stato successivo
		break;

		case (1):
			LcdEN=1;											// strobe		
			LcdByteStatus ++;								// passa allo stato successivo
			LcdEN=0;
		break;
				
		case (2):	
			LcdData=LcdChar; 								// lower nibble	
			I2c[LcdPtr].Flag.Tx = 1;					// in modalita' 4 bit trasmette  un
			I2c[LcdPtr].TxBuff[0] = LcdByteSet() ;	// nibble per volta
			LcdByteStatus ++;								// passa allo stato successivo
		break;

		case (3):
			LcdEN=1;											// strobe		
			LcdPutCharFlag = 0;	// l'invio del carattere e' finito, se la routine a
										// livello superiore ha altri byte da inviare
										// inizia una nuova sequenza con LcdPutChar
			LcdEN=0;
		break;
		
	} // end switch

} // Lcd4Bit
/**************************************************************************************/


/*Lcd8Bit *****************************************************************************
Invia un byte all'LCD nella modalita' a 8 bit tramite I2C, usata solo per l'init
L'I/O expander 1 usato per pilotare l'LCD e' il device I2c numero 3
Il driver LCD HD44780 con il clock interno a 270KHz richiede una pausa di 37uSec
tra un carattere e l'altro. Per inviare un byte tramite l'I/O expander PCF8574 con il
clock a 100KHz ci vogliono almeno 90uSec (8 bit + ACK * 10uSec a bit), piu' che 
sufficienti quindi per le temporizzazioni richieste.
*/

void Lcd8Bit (void)
{	
	switch (LcdByteStatus)
	{
		case (0):
			LcdEN=0;	
			LcdData=LcdChar;	
			I2c[LcdPtr].Flag.Tx = 1;					// un byte da trasmettere
			I2c[LcdPtr].TxBuff[0] = LcdByteSet() ;	// compone il byte da inviare e lo passa
																// alle routine I2C. Il buffer e' stato													// controllato prima e quindi e'
																// sicuramente vuoto
			LcdByteStatus ++;								// passa allo stato successivo
		break;

		case (1):
			LcdEN=1;											// strobe		
			LcdPutCharFlag = 0;	// l'invio del carattere e' finito, se la routine a
										// livello superiore ha altri byte da inviare
										// inizia una nuova sequenza con LcdPutChar
			LcdEN=0;;
		break;
		
	} // end switch
	
} // Lcd8Bit
/**************************************************************************************/


/*LcdInit *****************************************************************************
Per l'inizializzazione del display LCD 
Interfaccia a 4 bit, 2 linee di visualizzazione,e caratteri di 5 * 7 punti.
*/
void LcdInit (void)
{
 
	switch (LcdInitStatus)
	{
		case 0:
			LcdPutChar (0x03, 0, 8);					// invia carattere 0x30 all'LCD
			LcdInitStatus ++;								// passa allo stato successivo
			break;

		case 1:
			TimerLcd=5;										// inizializza timer 5mS
			LcdInitStatus ++;								// passa allo stato successivo
			break;
		
		case 2:
		
			LcdPutChar (0x03, 0, 8);					// invia carattere 0x30 all'LCD
			LcdInitStatus ++;								// passa allo stato successivo
			break;

		case 3:
			TimerLcd=5;										// inizializza timer 5mS
			LcdInitStatus ++;								// passa allo stato successivo
			break;

		case 4:
			LcdPutChar (0x03, 0, 8);					// invia carattere 0x30 all'LCD
			LcdInitStatus ++;								// passa allo stato successivo
			break;
		
		case 5:
			LcdPutChar (0x02, 0, 8);						// Modalita' 4 bit
			LcdInitStatus ++;								// passa allo stato successivo
			break;
		
		case 6:
			TimerLcd=5;										// inizializza timer 5mS
			LcdInitStatus ++;								// passa allo stato successivo
			break;

		case 7:
			LcdPutChar (0x28, 0, 4);					// 1/16 duty, 5x7 font
			LcdInitStatus ++;								// passa allo stato successivo
			break;

		case 8:
			LcdPutChar (0x0C, 0, 4);					// display on, blink cursor off
			LcdInitStatus ++;								// passa allo stato successivo
			break;

		case 9:
			LcdPutChar (0x06, 0, 4);					// entry mode=increment cursor position
			LcdInitStatus ++;								// passa allo stato successivo
			break;
			
		case 10:
			LcdPutChar (0x01, 0, 4);					// clear display
			LcdInitStatus ++;								// passa allo stato successivo
			break;

		case 11:
			TimerLcd=5;										// inizializza timer 5mS
			LcdInitStatus ++;								// passa allo stato successivo
			break;

		case 12:
			LcdBL=1;											// Backlight ON
			LcdPutStringInitRom(0,Ver);				// scrive versione
			LcdInitStatus ++;								// passa allo stato successivo
			break;

		case 13:
			LcdBL=0;											// Backlight OFF
			LedRossoON;										// rimane in loop ciclando il led
			LedGialloOFF;									// ogni 250mS
			LedVerdeOFF;
			LcdInitStatus ++;								// passa allo stato successivo
			LcdPutStringInitC2A(0x40,GasVal);		// visualizza valore sensore gas
			break;

		case 14:
			TimerLcd=250;									// inizializza timer 250mS
			LcdInitStatus ++;								// passa allo stato successivo
			break;
								
		case 15:
			LedRossoOFF;
			LedGialloON;
			LcdInitStatus ++;								// passa allo stato successivo
			break;

		case 16:
			TimerLcd=250;									// inizializza timer 250mS
			LcdInitStatus ++;								// passa allo stato successivo
			break;

		case 17:
			LedGialloOFF;
			LedVerdeON;			
			LcdInitStatus ++;								// passa allo stato successivo
			break;
				
		case 18:
			TimerLcd=250;									// inizializza timer 250mS
			LcdInitStatus = 13;							// ricomincia il ciclo dei led
			break;


		default:
			break;
		
	} // end switch	
	
} // LcdInit
/**************************************************************************************/

